﻿using Fleck;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Net.Sockets;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading;

namespace FleckProxy
{
    class Program
    {
        static Dictionary<Guid, SessionData> _Sessions = new Dictionary<Guid, SessionData>();

        static void Main(string[] args)
        {
            Console.Clear();

            // Override default FleckLog implementation
            FleckLog.LogAction = LogAction;

            if (ParseCommandLine(args))
            {
                // Bring requested servers online
                WebSocketServer WsListener = (Config.WsPort > 0) ? StartListener(Config.WsPort, "ws") : null;
                WebSocketServer WssListener = (Config.WssPort > 0) ? StartListener(Config.WssPort, "wss") : null;

                // Ensure requested servers came online
                if ((Config.WsPort > 0) && (WsListener == null)) {
                    Console.WriteLine();
                    Console.WriteLine("Error bringing ws:// server online.  Server will now shutdown");
                    Console.WriteLine();
                } else if ((Config.WssPort > 0) && (WssListener == null)) {
                    Console.WriteLine();
                    Console.WriteLine("Error bringing wss:// server online.  Server will now shutdown");
                    Console.WriteLine();
                }
                else
                {
                    // Wait for 'Q' to quit
                    Console.WriteLine();
                    Console.WriteLine("Hit 'Q' to quit");
                    Console.WriteLine();
                    while (true)
                    {
                        ConsoleKeyInfo CKI = Console.ReadKey(true);
                        if (CKI.Key == ConsoleKey.Q)
                        {
                            FleckLog.Info("Shutting down");
                            break;
                        }
                        else
                        {
                            // Don't use 100% CPU if we're running as a daemon
                            Thread.Sleep(100);
                        }
                    }
                }
            }

            if (Debugger.IsAttached)
            {
                Console.WriteLine();
                Console.WriteLine("DEBUG: Hit a key to quit");
                Console.ReadKey();
            }
        }

        private static void ClientSend(IWebSocketConnection clientConnection, string text)
        {
            if (clientConnection.ConnectionInfo.NegotiatedSubProtocol == "binary")
            {
                // This assumes we'll only ever be passing plain-text to this function
                clientConnection.Send(Encoding.ASCII.GetBytes(text));
            }
            else
            {
                clientConnection.Send(text);
            }
        }

        private static void ClientSend(IWebSocketConnection clientConnection, byte[] data)
        {
            if (clientConnection.ConnectionInfo.NegotiatedSubProtocol == "binary")
            {
                clientConnection.Send(data);
            }
            else
            {
                char[] chars = new char[data.Length];
                for (int i = 0; i < data.Length; i++)
                {
                    chars[i] = (char)data[i];
                }

                clientConnection.Send(new string(chars));
            }
        }

        private static void CloseClientConnection(Guid connectionId)
        {
            try
            {
                if (_Sessions.ContainsKey(connectionId))
                {
                    SessionData SD = _Sessions[connectionId];

                    if ((SD.ClientSocket != null) && SD.ClientSocket.IsAvailable)
                    {
                        FleckLog.Info(connectionId.ToString() + " Closing client connection");
                        SD.ClientSocket.OnClose = null; // OnClose calls CloseConnections, and we don't want to get in a loop
                        SD.ClientSocket.Close();
                        FleckLog.Info(connectionId.ToString() + " Client connection closed");
                    }

                    SD.ClientSocket = null;
                }
            }
            catch (SocketException sex)
            {
                FleckLog.Error(connectionId.ToString() + " Error closing client connection (SocketError=" + ((int)sex.SocketErrorCode).ToString() + "/" + sex.SocketErrorCode.ToString() + ": " + sex.Message + ")", sex);
            }
            catch (Exception ex)
            {
                FleckLog.Error(connectionId.ToString() + " Error closing client connection (Exception=" + ex.Message + ")", ex);
            }
        }

        private static void CloseConnections(Guid connectionId)
        {
            if (_Sessions.ContainsKey(connectionId))
            {
                SessionData SD = _Sessions[connectionId];
                if (!SD.Closing)
                {
                    SD.Closing = true;
                    CloseServerConnection(connectionId);
                    CloseClientConnection(connectionId);
                    _Sessions.Remove(connectionId);
                }
            }
        }

        private static void CloseServerConnection(Guid connectionId)
        {
            try
            {
                if (_Sessions.ContainsKey(connectionId))
                {
                    SessionData SD = _Sessions[connectionId];

                    if ((SD.ServerSocket != null) && SD.ServerSocket.Connected)
                    {
                        FleckLog.Info(connectionId.ToString() + " Closing server connection");
                        SD.ServerSocket.Shutdown(SocketShutdown.Receive);
                        SD.ServerSocket.Close(2);
                        SD.ServerSocket.Dispose();
                        FleckLog.Info(connectionId.ToString() + " Server connection closed");
                    }

                    SD.ServerSocket = null;
                }
            }
            catch (SocketException sex)
            {
                FleckLog.Error(connectionId.ToString() + " Error closing server connection (SocketError=" + ((int)sex.SocketErrorCode).ToString() + "/" + sex.SocketErrorCode.ToString() + ": " + sex.Message + ")", sex);
            }
            catch (Exception ex)
            {
                FleckLog.Error(connectionId.ToString() + " Error closing server connection (Exception=" + ex.Message + ")", ex);
            }
        }

        public static Action<LogLevel, string, Exception> LogAction = (level, message, ex) =>
        {
            if (level >= FleckLog.Level)
            {
                if (FleckLog.Level == LogLevel.Debug)
                {
                    Console.WriteLine("{0} [{1}] {2} {3}", DateTime.Now, level, message, ex);
                }
                else
                {
                    Console.WriteLine("{0} [{1}] {2}", DateTime.Now, level, message);
                }
            }
        };

        private static void OnClientBinaryMessage(IWebSocketConnection clientConnection, byte[] bytes)
        {
            try
            {
                if (bytes.Length > 0)
                {
                    if (_Sessions.ContainsKey(clientConnection.ConnectionInfo.Id))
                    {
                        SessionData SD = _Sessions[clientConnection.ConnectionInfo.Id];
                        SD.ServerSocket.Send(bytes);
                    }
                }
            }
            catch (SocketException sex)
            {
                FleckLog.Error(clientConnection.ConnectionInfo.Id.ToString() + " Error sending binary data from client to server (SocketError=" + ((int)sex.SocketErrorCode).ToString() + "/" + sex.SocketErrorCode.ToString() + ": " + sex.Message + ")", sex);
                CloseConnections(clientConnection.ConnectionInfo.Id);
            }
            catch (Exception ex)
            {
                FleckLog.Error(clientConnection.ConnectionInfo.Id.ToString() + " Error sending binary data from client to server (Exception=" + ex.Message + ")", ex);
                CloseConnections(clientConnection.ConnectionInfo.Id);
            }
        }

        /// <summary>
        /// Callback that handles the closing of the client connection
        /// </summary>
        /// <param name="clientConnection">The Fleck WebSocketConnection instance</param>
        private static void OnClientConnectionClosed(IWebSocketConnection clientConnection)
        {
            try
            {
                FleckLog.Info(clientConnection.ConnectionInfo.Id.ToString() + " Client connection closed");
                if (_Sessions.ContainsKey(clientConnection.ConnectionInfo.Id))
                {
                    // null the client connection, so further down when we call CloseConnections we don't try to close the client connection again
                    SessionData SD = _Sessions[clientConnection.ConnectionInfo.Id];
                    SD.ClientSocket = null;
                }
            }
            catch (Exception ex)
            {
                FleckLog.Error(clientConnection.ConnectionInfo.Id.ToString() + " Error handling client close (Exception=" + ex.Message + ")", ex);
            }
            finally
            {
                CloseConnections(clientConnection.ConnectionInfo.Id);
            }
        }

        /// <summary>
        /// Callback that handles the opening of the client connection
        /// </summary>
        /// <param name="clientConnection">The Fleck WebSocketConnection instance</param>
        private static void OnClientConnectionOpened(IWebSocketConnection clientConnection, string scheme)
        {
            try
            {
                FleckLog.Info(clientConnection.ConnectionInfo.Id.ToString() + " Accepted " + scheme + " connection from " + clientConnection.ConnectionInfo.ClientIpAddress + ":" + clientConnection.ConnectionInfo.ClientPort.ToString() + " for " + clientConnection.ConnectionInfo.Path + " (Protocol=" + clientConnection.ConnectionInfo.NegotiatedSubProtocol + ")");

                // Defaults for redirect location
                int Port = Config.TargetPort;
                string Hostname = Config.TargetHostname;

                // Check if we should override the defaults with user selected values
                if (!string.IsNullOrEmpty(Config.RelayFilename))
                {
                    bool CanRelay = false;

                    string[] HostAndPort = clientConnection.ConnectionInfo.Path.Split('/');
                    if ((HostAndPort.Length == 3) && (int.TryParse(HostAndPort[2], out Port)))
                    {
                        Hostname = HostAndPort[1];

                        if (File.Exists(Config.RelayFilename))
                        {
                            try
                            {
                                string[] AllowedHosts = File.ReadAllLines(Config.RelayFilename);
                                if (AllowedHosts.Length > 0)
                                {
                                    string[] AllowedPorts = AllowedHosts[0].Split(',');
                                    if (AllowedPorts.Contains(Port.ToString()))
                                    {
                                        CanRelay = true;
                                    }
                                    else
                                    {
                                        string RequestedHostPort = Hostname + ":" + Port.ToString();
                                        foreach (string AllowedHost in AllowedHosts)
                                        {
                                            if (AllowedHost == RequestedHostPort)
                                            {
                                                CanRelay = true;
                                                break;
                                            }
                                        }
                                    }
                                }
                            }
                            catch (Exception ex)
                            {
                                FleckLog.Error(clientConnection.ConnectionInfo.Id.ToString() + " Error reading relay file '" + Config.RelayFilename + "' (Exception=" + ex.Message + ")", ex);
                            }
                        }
                    }

                    if (!CanRelay)
                    {
                        FleckLog.Warn(clientConnection.ConnectionInfo.Id.ToString() + " Rejecting request for " + Hostname + ":" + Port.ToString());
                        ClientSend(clientConnection, "Sorry, for security reasons this proxy won't connect to " + Hostname + ":" + Port.ToString() + "\r\n");
                        Thread.Sleep(2500);
                        clientConnection.Close();
                        return;
                    }
                }

                // Connect to remote server
                ClientSend(clientConnection, "Connecting to " + Hostname + ":" + Port.ToString() + "...\r\n");
                Socket ServerSocket = OpenServerConnection(Hostname, Port, clientConnection.ConnectionInfo.Id);
                if (ServerSocket == null)
                {
                    FleckLog.Warn(clientConnection.ConnectionInfo.Id.ToString() + " Unable to connect to " + Hostname + ":" + Port.ToString());
                    ClientSend(clientConnection, "Sorry, I wasn't able to connect to " + Hostname + ":" + Port.ToString() + "\r\n");
                    Thread.Sleep(2500);
                    clientConnection.Close();
                    return;
                }
                else
                {
                    // Create new session data instance to hold client and server connections
                    SessionData Session = new SessionData(clientConnection, ServerSocket);
                    _Sessions.Add(clientConnection.ConnectionInfo.Id, Session);

                    // Start listening for data from server
                    ServerSocket.BeginReceive(Session.ReceiveBuffer, 0, Session.ReceiveBuffer.Length, SocketFlags.None, new AsyncCallback(OnServerReceive), clientConnection.ConnectionInfo.Id);
                }
            }
            catch (SocketException sex)
            {
                FleckLog.Error(clientConnection.ConnectionInfo.Id.ToString() + " Error handling client open (SocketError=" + ((int)sex.SocketErrorCode).ToString() + "/" + sex.SocketErrorCode.ToString() + ": " + sex.Message + ")", sex);
            }
            catch (Exception ex)
            {
                FleckLog.Error(clientConnection.ConnectionInfo.Id.ToString() + " Error handling client open (Exception=" + ex.Message + ")", ex);
            }
        }

        private static void OnClientTextMessage(IWebSocketConnection clientConnection, string message)
        {
            if (message.Length > 0)
            {
                if (_Sessions.ContainsKey(clientConnection.ConnectionInfo.Id))
                {
                    SessionData SD = _Sessions[clientConnection.ConnectionInfo.Id];

                    try
                    {
                        byte[] bytes = new byte[message.Length];
                        for (int i = 0; i < message.Length; i++)
                        {
                            bytes[i] = (byte)message[i];
                        }

                        SD.ServerSocket.Send(bytes);
                    }
                    catch (SocketException sex)
                    {
                        FleckLog.Error(clientConnection.ConnectionInfo.Id.ToString() + " Error sending text message from client to server (SocketError=" + ((int)sex.SocketErrorCode).ToString() + "/" + sex.SocketErrorCode.ToString() + ": " + sex.Message + ")", sex);
                        CloseConnections(clientConnection.ConnectionInfo.Id);
                    }
                    catch (Exception ex)
                    {
                        FleckLog.Error(clientConnection.ConnectionInfo.Id.ToString() + " Error sending text message from client to server (Exception=" + ex.Message + ")", ex);
                        CloseConnections(clientConnection.ConnectionInfo.Id);
                    }
                }
            }
        }

        /// <summary>
        /// Callback that receives data from the server and sends to the client
        /// </summary>
        /// <param name="ar">The Guid for this session</param>
        private static void OnServerReceive(IAsyncResult ar)
        {
            Guid ConnectionId = (Guid)ar.AsyncState;

            try
            {
                if (_Sessions.ContainsKey(ConnectionId))
                {
                    SessionData SD = _Sessions[ConnectionId];

                    if (SD.ServerSocket != null)
                    {
                        int BytesRead = SD.ServerSocket.EndReceive(ar);
                        if (BytesRead == 0)
                        {
                            // Server closed connection
                            FleckLog.Info(ConnectionId.ToString() + " Server connection closed");
                            SD.ServerSocket = null;
                            CloseConnections(ConnectionId);
                        }
                        else
                        {
                            if (SD.ClientSocket.IsAvailable)
                            {
                                // Send data from server to client
                                ClientSend(SD.ClientSocket, SD.ReceiveBuffer.Take(BytesRead).ToArray());

                                // Start listening for data from server
                                SD.ServerSocket.BeginReceive(SD.ReceiveBuffer, 0, SD.ReceiveBuffer.Length, SocketFlags.None, new AsyncCallback(OnServerReceive), ConnectionId);
                            }
                            else
                            {
                                CloseConnections(ConnectionId);
                            }
                        }
                    }
                }
            }
            catch (ObjectDisposedException)
            {
                if (_Sessions.ContainsKey(ConnectionId))
                {
                    FleckLog.Info(ConnectionId.ToString() + " Server connection closed");
                    SessionData SD = _Sessions[ConnectionId];
                    SD.ServerSocket = null;
                }
                CloseConnections(ConnectionId);
            }
            catch (SocketException sex)
            {
                FleckLog.Error(ConnectionId.ToString() + " Error sending data from server to client (SocketError=" + ((int)sex.SocketErrorCode).ToString() + "/" + sex.SocketErrorCode.ToString() + ": " + sex.Message + ")", sex);
                CloseConnections(ConnectionId);
            }
            catch (Exception ex)
            {
                FleckLog.Error(ConnectionId.ToString() + " Error sending data from server to client (Exception=" + ex.Message + ")", ex);
                CloseConnections(ConnectionId);
            }
        }

        /// <summary>
        /// Connect a websocket client to a remote server
        /// </summary>
        /// <param name="hostname">The hostname of the remote server to connect to</param>
        /// <param name="port">The port of the remote server to connect to</param>
        /// <returns>The server socket if the connect was successfull, or null if the connect failed</returns>
        private static Socket OpenServerConnection(string hostname, int port, Guid connectionId)
        {
            try
            {
                Socket ServerSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                ServerSocket.Blocking = true;
                ServerSocket.Connect(hostname, port);
                if (ServerSocket.Connected)
                {
                    return ServerSocket;
                }
            }
            catch (SocketException sex)
            {
                FleckLog.Error(connectionId.ToString() + " Error connecting to " + hostname + ":" + port.ToString() + " (SocketError=" + ((int)sex.SocketErrorCode).ToString() + "/" + sex.SocketErrorCode.ToString() + ": " + sex.Message + ")", sex);
            }
            catch (Exception ex)
            {
                FleckLog.Error(connectionId.ToString() + " Error connecting to " + hostname + ":" + port.ToString() + " (Exception=" + ex.Message + ")", ex);
            }

            // If we get here the connect failed
            return null;
        }

        private static bool ParseCommandLine(string[] args)
        {
            for (int i = 0; i < args.Length; i++)
            {
                switch (args[i])
                {
                    case "-c":
                    case "--cert":
                        i += 1;
                        Config.CertFilename = args[i];
                        break;
                    case "-?":
                    case "-h":
                    case "--help":
                        ShowHelp();
                        return false;
                    case "-l":
                    case "--loglevel":
                        i += 1;
                        try
                        {
                            FleckLog.Level = (Fleck.LogLevel)Enum.Parse(typeof(Fleck.LogLevel), args[i]);
                        }
                        catch (Exception)
                        {
                            Console.WriteLine();
                            Console.WriteLine("Invalid loglevel '" + args[i] + "'");
                            Console.WriteLine();
                            return false;
                        }
                        break;
                    case "-pw":
                    case "--password":
                        i += 1;
                        Config.CertPassword = args[i];
                        break;
                    case "-ws":
                    case "--ws":
                        i += 1;
                        try
                        {
                            Config.WsPort = Convert.ToInt16(args[i]);
                        }
                        catch (Exception)
                        {
                            Console.WriteLine();
                            Console.WriteLine("Invalid port '" + args[i] + "'");
                            Console.WriteLine();
                            return false;
                        }
                        break;
                    case "-wss":
                    case "--wss":
                        i += 1;
                        try
                        {
                            Config.WssPort = Convert.ToInt16(args[i]);
                        }
                        catch (Exception)
                        {
                            Console.WriteLine();
                            Console.WriteLine("Invalid port '" + args[i] + "'");
                            Console.WriteLine();
                            return false;
                        }
                        break;
                    case "-r":
                    case "--relay":
                        i += 1;
                        Config.RelayFilename = args[i];
                        break;
                    case "-t":
                    case "--target":
                        i += 1;
                        if (args[i].Contains(":"))
                        {
                            Config.TargetHostname = args[i].Split(':')[0];
                            try
                            {
                                Config.TargetPort = Convert.ToInt16(args[i].Split(':')[1]);
                            }
                            catch (Exception)
                            {
                                Console.WriteLine();
                                Console.WriteLine("Invalid target port '" + args[i].Split(':')[1] + "'");
                                Console.WriteLine();
                                return false;
                            }
                        }
                        else
                        {
                            Config.TargetHostname = args[i];
                        }
                        break;
                    default:
                        ShowHelp();
                        Console.WriteLine();
                        Console.WriteLine("Error:");
                        Console.WriteLine();
                        Console.WriteLine("  Unknown parameter '" + args[i] + "'");
                        Console.WriteLine();
                        return false;
                }
            }

            bool Result = false;

            if (Config.WsPort > 0)
            {
                Result = true;
            }
            
            if (Config.WssPort > 0)
            {
                if (File.Exists(Config.CertFilename))
                {
                    Result = true;
                }
                else
                {
                    ShowHelp();
                    Console.WriteLine();
                    Console.WriteLine("Error:");
                    Console.WriteLine();
                    Console.WriteLine("  Passing --wss also requires passing --cert");
                    Console.WriteLine();
                    return false;
                }
            }


            if (!Result)
            {
                ShowHelp();
                Console.WriteLine();
                Console.WriteLine("Error:");
                Console.WriteLine();
                Console.WriteLine("  Either (or both) --ws or --wss must be passed");
                Console.WriteLine();
                return false;
            }

            return Result;
        }

        private static void ShowHelp()
        {
            Console.WriteLine();
            Console.WriteLine("Usage: FleckProxy.exe [options]");
            Console.WriteLine();
            Console.WriteLine("Options:");
            Console.WriteLine();
            Console.WriteLine("  -ws <port>                 Port to listen for ws:// connections on");
            Console.WriteLine("  --ws <port>                Not supplying will disable ws:// connections");
            Console.WriteLine();
            Console.WriteLine("  -t <host:port>             Telnet server to redirect to");
            Console.WriteLine("  --target <host:port>       Default is localhost:23");
            Console.WriteLine();
            Console.WriteLine("  -wss <port>                Port to listen for wss:// connections on");
            Console.WriteLine("  --wss <port>               Not supplying will disable wss:// connections");
            Console.WriteLine();
            Console.WriteLine("  -c <filename>              PKCS12 file containing private key and cert chain");
            Console.WriteLine("  --cert <filename>          Only needed if your site uses https://");
            Console.WriteLine();
            Console.WriteLine("  -pw <password>             Password to use to open the PKCS12 file");
            Console.WriteLine("  --password <password>      Only needed if PKCS12 file is password protected");
            Console.WriteLine();
            Console.WriteLine("  -l <level>                 Set log level (Debug, Info, Warn, Error)");
            Console.WriteLine("  --loglevel <level>         Default is Info");
            Console.WriteLine();
            Console.WriteLine("  -?, -h, --help             Display this screen");
            Console.WriteLine();
            Console.WriteLine("Notes:");
            Console.WriteLine();
            Console.WriteLine("- You must pass a port via either --ws (plain ws:// connection) or");
            Console.WriteLine("  --wss (ssl wss:// connection) to start the server.  You can also pass both");
            Console.WriteLine("  plain and ssl ports by using --ws and --wss at the same time");
            Console.WriteLine();
            Console.WriteLine("- If you pass a port via --wss, then you must also pass a PKCS12 certificate");
            Console.WriteLine("  via the --cert parameter.  If the certificate is password protected, then");
            Console.WriteLine("  you must also pass the --password parameter.  Note the password may be");
            Console.WriteLine("  visible to other users of the system, so you may be better off with a");
            Console.WriteLine("  password-less certificate that is secured so only you have access to read it");
            //Console.WriteLine("345678901234567890123456789012345678901234567890123456789012345678901234567890");
        }

        /// <summary>
        /// Start the listener
        /// </summary>
        private static WebSocketServer StartListener(int port, string scheme)
        {
            try
            {
                // Output settings to screen
                FleckLog.Info("Starting server on port " + port);
                FleckLog.Info("Redirecting to " + Config.TargetHostname + ":" + Config.TargetPort.ToString());
                if (!string.IsNullOrEmpty(Config.RelayFilename))
                {
                    FleckLog.Info("Relaying according to rules in '" + Config.RelayFilename + "'");
                }
                Console.WriteLine();

                // Create listener
                var Listener = new WebSocketServer(scheme + "://0.0.0.0:" + port);
                Listener.SupportedSubProtocols = new string[] { "binary", "plain" };

                // Try to load certificate, if necessary
                if (scheme == "wss")
                {
                    if (string.IsNullOrEmpty(Config.CertFilename))
                    {
                        FleckLog.Error("Error loading certificate (filename not supplied)");
                        return null;
                    }
                    else
                    {
                        try
                        {
                            Listener.Certificate = new X509Certificate2(Config.CertFilename, Config.CertPassword);
                        }
                        catch (CryptographicException cex)
                        {
                            FleckLog.Error("Error loading certificate '" + Config.CertFilename + "' (Exception=" + cex.Message + ")", cex);
                            return null;
                        }
                        catch (Exception ex)
                        {
                            FleckLog.Error("Error loading certificate '" + Config.CertFilename + "' (Exception=" + ex.Message + ")", ex);
                            return null;
                        }
                    }
                }

                // Start listener
                Listener.Start(clientConnection =>
                {
                    clientConnection.OnOpen = () => { OnClientConnectionOpened(clientConnection, scheme); };

                    clientConnection.OnClose = () => { OnClientConnectionClosed(clientConnection); };

                    clientConnection.OnMessage = message => { OnClientTextMessage(clientConnection, message); };

                    clientConnection.OnBinary = bytes => { OnClientBinaryMessage(clientConnection, bytes); };
                });

                return Listener;
            }
            catch (SocketException sex)
            {
                FleckLog.Error("Unable to start server (SocketError=" + ((int)sex.SocketErrorCode).ToString() + "/" + sex.SocketErrorCode.ToString() + ": " + sex.Message + ")", sex);
                return null;
            }
            catch (Exception ex)
            {
                FleckLog.Error("Unable to start server (Exception=" + ex.Message + ")", ex);
                return null;
            }
        }
    }
}
